#!/usr/bin/env python3

import unittest
import os
import signal
import subprocess
import json
import uuid
import random
import string
import shlex

from glance_sync_slave import ImageSyncSlave

# This script tests the glance image sync script.
# It is meant to run during user acceptance tests and will create and remove
# images in glance. Since glance does not fully remove images from its database
# on image deletion you will need to use the database cleanup script
# after user acceptances tests have been completed.
# It's located in the script_dir (default: /srv/glance_sync_slave/scripts) and
# will have to be copied to a glance unit to be run.


class TestSlaveSync(unittest.TestCase):

    def setUp(self):
        self.data_dir = '/srv/glance_sync_slave/data/'
        # take the same directory as source and storage data_dir, test cases
        # will add and remove files there according to their requirements
        self.source = '/srv/glance_sync_slave/data/'
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

        self.lockfile = '/tmp/glance_sync_slave.lock'
        self.slave = ImageSyncSlave()

    def test_script_can_connect_to_glance(self):
        """ Test if sync script can connect to glance """
        glance = self.slave.glance_connect()
        glance.images.list()

    def test_script_uploads_images(self):
        (" Test if sync script uploads image found in sync directory to glance")
        # create an iso with correct metadata
        random_id = str(uuid.uuid4())
        self.write_random_image(self.data_dir, random_id)

        # read metadata from disk
        meta_file = os.path.join(self.data_dir, random_id + '.json')
        with open(meta_file, 'r') as meta:
            file_metadata = json.load(meta)

        # trigger a sync
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # check if change is reflected in glance
        image = self.slave.glance.images.get(random_id)

        self.assertTrue(image.name == file_metadata['name'])

        # cleanup
        self.cleanup_temp_image(random_id)
        self.slave.main(self.data_dir, self.source, self.lockfile)

    def test_script_adds_metadata(self):
        """ Test if sync script adds image metadata to slave glance"""
        # create an iso with correct metadata
        random_id = str(uuid.uuid4())
        self.write_random_image(self.data_dir, random_id)

        # trigger a sync
        self.slave.main(self.data_dir, self.source, self.lockfile)

        self.cleanup_temp_image(random_id)

        # check if metadata is correct in glance
        image = self.slave.glance.images.get(random_id)
        self.assertTrue(
            image.name == 'canonical-user-acceptance-test-random-image')

    def test_script_updates_metadata(self):
        (" Test if sync script updates image metadata in glance when "
         "it changed in sync directory")
        # create an iso with correct metadata
        random_id = str(uuid.uuid4())
        self.write_random_image(self.data_dir, random_id)

        # trigger a sync
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # change metadata file in data_dir
        newname = "canonical-user-acceptance-test-metadata-changed"

        meta_file = os.path.join(self.data_dir, random_id + '.json')

        with open(meta_file, 'r') as meta:
            metadata = json.load(meta)

        metadata['name'] = newname

        with open(meta_file, 'w') as meta:
            json.dump(metadata, meta)

        # call glance_upload again
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # check if change is reflected in glance
        image = self.slave.glance.images.get(random_id)

        self.assertTrue(image.name == newname)

        # cleanup
        self.cleanup_temp_image(random_id)
        self.slave.main(self.data_dir, self.source, self.lockfile)

    def test_image_removal(self):
        ("Test if sync script removes images from slave glance that have "
         "been deleted in the sync directory")
        image_id = str(uuid.uuid4())
        self.write_random_image(self.data_dir, image_id)

        # trigger a sync
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # check glance for the new image
        glance_ids = [image.id for image in self.slave.glance.images.list()]
        self.assertTrue(image_id in glance_ids)

        os.remove(os.path.join(self.data_dir, image_id))
        os.remove(os.path.join(self.data_dir, image_id+'.json'))

        # trigger a sync
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # check if glance removed it
        glance_ids = [image.id for image in self.slave.glance.images.list()]

        self.cleanup_temp_image(image_id)
        self.assertTrue(image_id not in glance_ids)

    def test_glance_upload_catches_bad_checksum(self):
        """ Test if a checksum mismatch gets caught by the sync script"""

        image_id = str(uuid.uuid4())
        checksum = ''.join(random.choice(string.hexdigits) for
                           n in range(32)).lower()

        # create metadata
        test_metadata = {'id': image_id,
                         'name': 'canonical-user-acceptance-test-bad-checksum',
                         'checksum': checksum,
                         'disk_format': 'qcow2'}

        meta_file = os.path.join(self.data_dir,
                                 "{}.json".format(image_id))

        with open(meta_file, 'w') as dumpfile:
            json.dump(test_metadata, dumpfile)

        # generate a random iso, the checksum defined above will not match
        test_iso = os.path.join(self.data_dir, image_id)
        with open(test_iso, 'wb') as test_image:
            test_image.write(os.urandom(1024))

        # trigger a sync
        self.slave.main(self.data_dir, self.source, self.lockfile)

        glance_ids = [image.id for image in self.slave.glance.images.list()]

        self.assertTrue(image_id not in glance_ids)

        # cleanup
        os.remove(meta_file)
        os.remove(test_iso)

    def test_script_syncs_removed_metadata(self):
        ("Test if sync script removes properties from slave glance that were "
         "removed in the sync directory json file")
        # create an iso with correct metadata
        random_id = str(uuid.uuid4())
        self.write_random_image(self.data_dir, random_id)

        meta_file = os.path.join(self.data_dir, random_id + '.json')
        with open(meta_file, 'r') as meta:
            metadata = json.load(meta)

        # add a custom field
        field_name = 'canonical-test-field'
        field_value = 'user acceptance test'
        metadata[field_name] = field_value
        with open(meta_file, 'w') as meta:
            json.dump(metadata, meta)

        # call glance_upload
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # make sure it got uploaded
        image = self.slave.glance.images.get(random_id)
        self.assertTrue(str(image[field_name]) == field_value)

        # remove a property from metadata
        del metadata[field_name]
        with open(meta_file, 'w') as meta:
            json.dump(metadata, meta)

        # call glance_upload again
        self.slave.main(self.data_dir, self.source, self.lockfile)

        # check if change is reflected in glance
        image = self.slave.glance.images.get(random_id)

        self.cleanup_temp_image(random_id)

        self.assertTrue(field_name not in image.keys())

    def test_glance_upload_respects_lockfile(self):
        """ Test if sync script respects the lockfile of another sync run """

        # make sure we're good to go
        self.assertFalse(self.slave.file_locked(self.lockfile))

        # start a sync in the background, copy from data_dir to data_dir just
        # to keep it busy for a bit
        command = "python glance_sync_slave.py -d {} -s {}".format(
            self.data_dir, self.data_dir)

        with open(os.devnull, "w") as DEVNULL:
            blocker = subprocess.Popen(shlex.split(command),
                                       stdout=DEVNULL,
                                       stderr=DEVNULL)

        blocker.send_signal(signal.SIGSTOP)

        # start a conflicting sync run
        try:
            subprocess.check_output(['python', 'glance_sync_slave.py',
                                     '-d', self.data_dir,
                                     '-s', self.data_dir])
        except subprocess.CalledProcessError as e:
            print(e.output)
            self.assertTrue("sync already in progress" in e.output)

        # let the blocker finish
        blocker.send_signal(signal.SIGCONT)
        blocker.wait()

    def write_random_image(self, data_dir, image_id):

        # generate a random iso and the corresponding json metadata file
        test_iso = os.path.join(data_dir, image_id)
        with open(test_iso, 'wb') as test_image:
            test_image.write(os.urandom(1024))

        checksum = self.slave.get_md5(test_iso)
        test_metadata = {'id': image_id,
                         'checksum': checksum,
                         'name': 'canonical-user-acceptance-test-random-image',
                         'disk_format': 'qcow2',
                         'container_format': 'bare'}

        meta_file = os.path.join(data_dir,
                                 "{}.json".format(image_id))
        with open(meta_file, 'w') as dumpfile:
            json.dump(test_metadata, dumpfile)

    def cleanup_temp_image(self, image_id):
        # cleanup
        test_image = os.path.join(self.data_dir, image_id)
        test_metafile = "{}.json".format(test_image)

        if os.path.exists(test_image):
            os.remove(test_image)
        if os.path.exists(test_metafile):
            os.remove(test_metafile)
