#!/usr/bin/env python

import os
import pwd
import grp
import base64
import shutil
import subprocess
from charmhelpers.core import hookenv, templating
from charmhelpers import fetch
from charms.reactive import hook, when, set_state, remove_state


@hook('install')
def install_glance_sync_slave():
    hookenv.status_set('maintenance', 'Installing')
    fetch.apt_update()
    fetch.apt_install('python-glanceclient')
    fetch.apt_install('python-nose')
    fetch.apt_install('python-dateutil')

    # configure the variables that have a default value
    configure_config_dir()
    configure_data_dir()
    configure_log_dir()
    configure_script_dir()

    install_sync_script()
    install_db_cleanup_script()

    homedir = os.path.expanduser('~ubuntu')
    ssh_identity = '{}/.ssh/id_rsa'.format(homedir)

    # make sure cron sync user has an ssh key for the sync, generate if
    # necessary
    if not os.path.exists(ssh_identity):
        command = ['ssh-keygen', '-t', 'rsa', '-N', '', '-f', ssh_identity]
        proc = subprocess.Popen(command,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate()
        if proc.returncode:
            print("ERROR: problem generating ssh key '{}':".format(command))
            print(stderr)
        os.chown(ssh_identity,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)

    hookenv.status_set('active', 'Installation complete')
    set_state('glance-sync-slave.installed')


@hook('upgrade-charm')
def upgrade_glance_sync_slave():
    install_glance_sync_slave()
    configure_novarc()


@when('config.changed.config_dir')
def configure_config_dir():
    config_dir = hookenv.config('config_dir')

    if not os.path.exists(config_dir):
        os.makedirs(config_dir)
        os.chown(config_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
    configure_cron()


@when('config.changed.data_dir')
def configure_data_dir():
    data_dir = hookenv.config('data_dir')
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
        os.chown(data_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
    configure_cron()


@when('config.changed.log_dir')
def configure_log_dir():
    log_dir = hookenv.config('log_dir')
    hookenv.status_set('maintenance', 'setting up logrotate')
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
        os.chown(log_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)

    templating.render(source='logrotate.d.j2',
                      target='/etc/logrotate.d/glance_sync_slave',
                      owner='root',
                      perms=0o644,
                      context=hookenv.config())
    configure_cron()


@when('config.changed.script_dir')
def configure_script_dir():
    script_dir = hookenv.config('script_dir')
    if not os.path.exists(script_dir):
        os.makedirs(script_dir)
        os.chown(script_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
    install_sync_script()
    install_db_cleanup_script()
    configure_cron()


def install_sync_script():
    # copy sync script and user acceptance tests in place
    script_dir = hookenv.config('script_dir')
    files = ['glance_sync_slave.py',
             os.path.join('tests', 'test_glance_sync_slave.py'),
             os.path.join('tests', '__init__.py')]

    hookenv.log('installing slave sync script')
    for file in files:
        source = os.path.join(hookenv.charm_dir(), 'files', file)
        dest = os.path.join(script_dir, file)

        if not os.path.exists(os.path.dirname(dest)):
            os.makedirs(os.path.dirname(dest))

        shutil.copy(source, dest)
        os.chown(dest,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
        os.chmod(dest, 0o640)


def install_db_cleanup_script():
    hookenv.status_set('maintenance', 'Installing DB cleanup script')
    # copy sync script in place
    script_dir = hookenv.config('script_dir')
    source = os.path.join(hookenv.charm_dir(),
                          'files',
                          'db_purge_deleted_glance_images.py')
    dest = os.path.join(script_dir, 'db_purge_deleted_glance_images.py')
    shutil.copy(source, dest)
    os.chown(dest,
             pwd.getpwnam('ubuntu').pw_uid,
             grp.getgrnam('ubuntu').gr_gid)
    os.chmod(dest, 0o740)
    hookenv.status_set('active', 'DB cleanup script installed')


@when('config.changed.novarc')
def configure_custom_novarc():
    configure_novarc()

@when('config.changed.master_{password,auth_url,glance_url,region}')
def configure_master_novarc():
    context = {}
    for i in ('master_password', 'master_auth_url',
              'master_glance_url', 'master_region'):
        value = hookenv.config(i)
        if not value:
            hookenv.status_set('blocked', 'master service credentials '
                               'missing')
            return
        context[i] = value

    glancesync_novarc_file = os.path.join(config_dir, 'glancesync.novarc')

    templating.render(source='glancesync.novarc.j2',
                      target=glancesync_novarc_file,
                      owner='ubuntu',
                      perms=0o600,
                      context=relation)
    hookenv.status_set('active', 'glancesync.novarc configured')

@hook('{requires:keystone-admin}-relation-{joined,changed}')
def configure_relation_novarc(relation=None):
    configure_novarc()


# If we have a custom novarc config set it overrides any relation-pulled
# credentials. If the custom novarc goes away, configure_relation_novarc should
# fire.
def configure_novarc():
    hookenv.status_set('maintenance', 'Configuring novarc')

    keystone_relations = hookenv.relations_of_type('keystone-admin')
    config_dir = hookenv.config('config_dir')
    novarc_file = os.path.join(config_dir, "novarc")

    if len(hookenv.config('novarc')) > 0:
        write_custom_novarc(novarc_file)
    elif len(keystone_relations) > 0:
        write_relation_novarc(novarc_file)
    else:
        message = "Please set novarc config or add a keystone relation."
        hookenv.log("ERROR: {}".format(message))
        if os.path.isfile(novarc_file):
            os.remove(novarc_file)
            remove_state('novarc.configured')

        hookenv.status_set('blocked', message)


def write_relation_novarc(path, relation=None):
    hookenv.status_set('maintenance', 'Configuring novarc from keystone '
                                      'relation')
    hookenv.log('configuring novarc based on keystone relation')
    keystone = hookenv.relations_of_type('keystone-admin')

    if not relation:
        keystone = hookenv.relations_of_type('keystone-admin')
        if len(keystone) > 0:
            relation = keystone[0]

    if 'service_password' in relation and relation['service_password']:
        templating.render(source='novarc.j2',
                          target=path,
                          owner='ubuntu',
                          perms=0o600,
                          context=relation)
        hookenv.status_set('active', 'novarc configured')
    else:
        hookenv.status_set('maintenance', 'keystone relation incomplete')


def write_custom_novarc(path):
    hookenv.status_set('maintenance', 'Configuring custom novarc')

    hookenv.log('configuring custom novarc')

    if hookenv.config('novarc'):
        with open(path, "wb") as f:
            f.write(base64.b64decode(hookenv.config('novarc')))

        os.chown(path,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
        os.chmod(path, 0o600)

        set_state('novarc.configured')
        hookenv.status_set('active', 'novarc configured')
    hookenv.status_set('maintenance', 'no novarc provided (juju set '
                       'glance-sync-master novarc=<base64 novarc>)')


@when('config.changed.sync_source')
def configure_sync_source():
    configure_cron()


@when('config.changed.sync_enabled')
def configure_cron():
    hookenv.status_set('maintenance', 'enabling sync')
    sync_enabled = hookenv.config('sync_enabled')

    hookenv.log('configuring sync cronjob')
    cron_file = os.path.join('/etc/cron.d', 'glance_sync_slave')
    if not hookenv.config('sync_source'):
        message = "Please set a sync_source to configure crontab."
        hookenv.log(message)
        hookenv.status_set('blocked', message)
        return

    if sync_enabled:
        hookenv.log('adding cronjob')
        templating.render(source='glance_sync_slave_cron.j2',
                          target=cron_file,
                          owner='root',
                          perms=0o644,
                          context=hookenv.config())

        set_state('cron.configured')
    else:
        hookenv.log("removing cronjob")
        if os.path.isfile(cron_file):
            os.remove(cron_file)
            remove_state('cron.configured')
    hookenv.status_set('active', 'sync configured')


@hook('nrpe-external-master-relation-changed')
def setup_nrpe_checks(nagios):
    hookenv.status_set('maintenance', 'Configuring nrpe checks')
    config = hookenv.config()
    nagios.add_check(['/usr/lib/nagios/plugins/check_file_age', '-f',
                      '-w 14400', '-c 25200', config['log_dir'],
                      'glance_sync_slave.log'],
                     name="check_glance_sync_slave_log",
                     description="Verify age of last image sync from glance "
                                 "to disk",
                     context=config["nagios_context"],
                     unit=hookenv.local_unit())

    # copy nrpe check plugin for stale lockfiles in place
    script_dir = hookenv.config('script_dir')
    files = ['check_stale_lockfile.py']

    hookenv.log('installing stale lockfile nrpe plugin')
    for file in files:
        source = os.path.join(hookenv.charm_dir(), 'files', file)
        dest = os.path.join(script_dir, file)

        if not os.path.exists(os.path.dirname(dest)):
            os.makedirs(os.path.dirname(dest))

        shutil.copy(source, dest)
        os.chown(dest,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
        os.chmod(dest, 0o740)

    nrpe_plugin = os.path.join(script_dir, 'check_stale_lockfile.py')
    nagios.add_check([nrpe_plugin, '-f', '/tmp/glance_sync_slave.lock',
                      '-w 72000', '-c 14400'],
                     name="check_glance_sync_slave_lockfile",
                     description="Verify age of image sync lockfile",
                     context=config["nagios_context"],
                     unit=hookenv.local_unit())

    hookenv.status_set('active', 'nrpe checks configured')
